import logging
from math import ceil
from functools import lru_cache
from typing import Iterable
from collections.abc import Mapping, MutableSet, MutableSequence
from frozendict import frozendict
from concurrent.futures import ThreadPoolExecutor
from swarkn.helpers import swallow_exception_wrap

logger = logging.getLogger(__name__)


def freeze(x):
    if isinstance(x, Mapping):
        return frozendict({k: freeze(v) for k, v in x.items()})
    elif isinstance(x, MutableSequence):
        return tuple(freeze(v) for v in x)
    elif isinstance(x, MutableSet):
        return frozenset(freeze(v) for v in x)
    return x

def unfreeze(x):
    if isinstance(x, frozendict):
        return {k: unfreeze(v) for k, v in x.items()}
    elif isinstance(x, tuple):
        return [unfreeze(v) for v in x]
    elif isinstance(x, frozenset):
        return {unfreeze(v) for v in x}
    return x


@lru_cache()
def _get_pool():
    return ThreadPoolExecutor(10)

def prl_map(func: callable, iterable: Iterable, chunksize=None, pool=None, ignore_failures=False):
    pool = pool or _get_pool()
    chunksize = chunksize or ceil(len(iterable) / len(pool._pool))
    chunks = list(chunk(iterable, chunksize))
    logger.info(f'splitting {len(iterable)} elements into {len(chunks)} buckets of size {chunksize}')
    wrapped_func = swallow_exception_wrap(func) if ignore_failures else func
    res = list(pool.map(wrapped_func, chunks))
    return res


import yaml
from functools import partial
from typing import Callable
from swarkn.collections import freeze, unfreeze
from cachetools.keys import _HashedTuple, _kwmark


def _freeze_args(freezefn: Callable, *args, **kwargs) -> _HashedTuple:
    args_ = tuple(freezefn(a) for a in args)
    if kwargs:
        kwargs_ = sum(sorted([
            (k, freezefn(v)) for k, v in kwargs.items()
        ]), _kwmark)
        return _HashedTuple(args_ + kwargs_)
    return _HashedTuple(args_)

freeze_args = partial(_freeze_args, freeze)
freeze_args_yaml = partial(_freeze_args, yaml.dump)

try:
    from freezedata import freeze_data
    freeze_args_obj = partial(_freeze_args, freeze_data)
except:
     pass





from cachetools import cached
from swarkn.caching import freeze_args, freeze_args_yaml

@cached({}, key=freeze_args) #freeze_args freeze_args_yaml
def cached_func(a, b=[1, 2, 4]):
    print(f'this only appears if function executed {a} {b}')
    return a, b


def test_freeze_args():
    kwarg = {'a': 2, 'b': {'c': [1, 2, 3], 'd': 1}, 'e': 'ee'}
    cached_func([1, 2, 3])
    cached_func(1)
    cached_func(1, [2, 3, 4])
    cached_func(1, b=kwarg)

    cached_func([1, 2, 3])
    cached_func(1)
    cached_func(1, [2, 3, 4])
    res = cached_func(1, b=kwarg)

if __name__ == '__main__':
    test_freeze_args()


from swarkn.helpers import swallow_exception, timer
def test_context_managers():
    with swallow_exception():
        raise RuntimeError('this is a test')

    with timer():
        x = 1

    with timer(None, logger, 'debug'):
        x = 1



from swarkn.collections import freeze, unfreeze
def test_freeze():
    f1 = freeze({'a': 1})
    f2 = freeze({1})
    f3 = freeze([1,2,3])
    f4 = freeze((1, 2, 3))
    f5 = freeze('this is a string')
    f6 = freeze({'a': [1, 2, 3], 'b': {'c': {1,2,3}, 'd': 'dd'}, 'e': 1})

    u1 = unfreeze(f1)
    u2 = unfreeze(f2)
    u3 = unfreeze(f3)
    u4 = unfreeze(f4)
    u5 = unfreeze(f5)
    u6 = unfreeze(f6)
